package xic.analysis;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import xic.InternalCompilerException;
import xic.ir.IRNode;
import xic.ir.IRPrinter;


public class CFGNode implements Comparable<CFGNode> {
	private List<CFGNode> inEdges;
	private List<CFGNode> outEdges;
	private Object value;

	private long id = 0;
	private static long id_counter = 0;
	private static Object lock = new Object();
	private int traversalIndex = 0;

	public CFGNode() {
		inEdges = new ArrayList<CFGNode>();
		outEdges = new ArrayList<CFGNode>();
		synchronized (lock) {
			id = id_counter++;
		}
	}

	public CFGNode(Object value) {
		this.value = value;
		inEdges = new ArrayList<CFGNode>();
		outEdges = new ArrayList<CFGNode>();
		synchronized (lock) {
			id = id_counter++;
		}
	}
	
	public void setTraversalIndex(int traversalIndex) {
		this.traversalIndex = traversalIndex;
	}

	public void flatten(Set<CFGNode> set) {	
		if (set.contains(this))
			return;
		set.add(this);
		
		for (CFGNode n: outEdges)
			n.flatten(set);
	}

	public void reverseFlatten(Set<CFGNode> set) {
		if (set.contains(this))
			return;
		set.add(this);
		for (CFGNode n: inEdges)
			n.reverseFlatten(set);
	}
	

	public Set<CFGNode> flatten() {
		Set<CFGNode> set = new HashSet<CFGNode>();
		flatten(set);
		return set;
	}

	public Set<CFGNode> reverseFlatten() {
		Set<CFGNode> set = new HashSet<CFGNode>();
		reverseFlatten(set);
		return set;
	}

	public void addInEdge(CFGNode source) {
		if (inEdges.contains(source))
			throw new InternalCompilerException("Internal error while building CFG");
		inEdges.add(source);
	}

	public void addOutEdge(CFGNode target) {
		if (outEdges.contains(target))
			throw new InternalCompilerException("Internal error while building CFG");
		outEdges.add(target);
	}
	
	public void addEdgeTo(CFGNode target) {
		addOutEdge(target);
		target.addInEdge(this);
	}

	public void addEdgeFrom(CFGNode source) {
		addInEdge(source);
		source.addOutEdge(this);
	}

	public void removeEdgeTo(CFGNode target) {
		removeOutEdge(target);
		target.removeInEdge(this);
	}

	public void removeEdgeFrom(CFGNode source) {
		removeInEdge(source);
		source.removeOutEdge(this);
	}
	
	
	public List<CFGNode> outEdges() {
		return outEdges;
	}
	
	public List<CFGNode> inEdges() {
		return inEdges;
	}

	public void setValue(Object value) {
		this.value = value;		
	}
	
	public Object value() {
		return value;
	}

	public String toString() {
		if (value instanceof IRNode)
			return IRPrinter.IRToString((IRNode) value, 30);
		else if( value == null) 
			return "";
		return value.toString();
	}

	public void removeOutEdge(CFGNode n) {
		if (!outEdges.contains(n))
			throw new InternalCompilerException("Tried to remove an out-edge to a nonexistent node!");
		outEdges.remove(n);
	}

	public void removeInEdge(CFGNode n) {
		if (!inEdges.contains(n))
			throw new InternalCompilerException("Tried to remove an in-edge to a nonexistent node!");
		inEdges.remove(n);	
	}

	public int compareTo(CFGNode o) {
		if (traversalIndex > o.traversalIndex)
			return -1;
		else if (traversalIndex < o.traversalIndex)
			return 1;
		else
			return 0;
	}
	
	public int hashCode() {
		// Want reference equality without non-determinism
		// introduced by Object.hashCode()
		return (int) id;
	}
	
	public boolean equals(Object obj) {
		if (!(obj instanceof CFGNode))
			return false;
		return ((CFGNode) obj).id == id;
	}

	public void removeSelf() {
		if (inEdges.size() > 1)
			throw new InternalCompilerException(
					"removeSelf(): unsupported for >1 incident edges");
		if (outEdges.size() > 1)
			throw new InternalCompilerException(
					"removeSelf(): unsupported for >1 outgoing edges");

		if (inEdges.size() == 1) {
			CFGNode pred = inEdges.get(0);
			pred.removeEdgeTo(this);
			if (outEdges.size() == 1) {
				CFGNode succ = outEdges.get(0);
				pred.addEdgeTo(succ);
			}
		}
		if (outEdges.size() == 1) {
			CFGNode succ = outEdges.get(0);
			removeEdgeTo(succ);
		}
	}
}
