package xic.analysis;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import xic.ErrorSink;
import xic.InternalCompilerException;
import xic.Statistics;
import xic.analysis.PropagatedConstants.Constant;
import xic.ast.ErrorReporter;
import xic.ir.IRCJump;
import xic.ir.IRConst;
import xic.ir.IRExpr;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRName;
import xic.ir.IRNode;
import xic.ir.IRStmt;
import xic.ir.IRTemp;
import xic.ir.StructHashSet.StructWrapper;
import xic.visitors.IRConstantFolder;
import xic.visitors.RewritingVisitor;


public class ConstPropagation {
	public static void transform (Statistics stats, int pass, Map<String, CFGNode> cfgs) {
		Map<StructWrapper, IRExpr> replaceMap = new HashMap<StructWrapper, IRExpr>();
		ErrorSink errorSink = new ErrorSink();
		RewritingVisitor rewritingVisitor = new RewritingVisitor(replaceMap);
		IRConstantFolder cfVisitor = new IRConstantFolder(errorSink);

		for (String funcName : cfgs.keySet()) {
			CFGNode head = cfgs.get(funcName);
			Set<CFGNode> nodeSet = head.flatten();
			PropagatedConstants cp = new PropagatedConstants(nodeSet);
			cp.solve();
			int numReplacements = 0;
			for (CFGNode node:nodeSet) {
				replaceMap.clear();
				if (node.value() instanceof IRStmt) {
					Map<String,Constant> valueMap = cp.in(node).values();
					for (String s : valueMap.keySet()) {
						Constant eval = valueMap.get(s);
						if (!eval.isBottom() && !eval.isTop()) {
							if (eval.isName())
								replaceMap.put(new StructWrapper(new IRTemp(s)), new IRName(eval.name()));
							else
								replaceMap.put(new StructWrapper(new IRTemp(s)), new IRConst(eval.value()));
						}
					}
					rewritingVisitor.setKnownExpressions(replaceMap);
					
					IRNode backup = ((IRNode) node.value());
					IRNode rewrittenNode = (IRNode) rewritingVisitor.visit((IRStmt) node.value());
					IRNode newNode = (IRNode) cfVisitor.visit(rewrittenNode);
					numReplacements += rewritingVisitor.getNumRewrites();

					cleanupIfCJump(node, backup, newNode);

					node.setValue(newNode);
				}
			}

			if (numReplacements > 0) {
				CFGGenerator.removeOrphans(head);
				stats.log("cp", "Pass " + pass + ": '" + funcName
						+ "': propagated " + numReplacements + " constants");
			}
		}
		

		if (errorSink.hasErrors()) {
			try {
				ErrorReporter.printErrors(errorSink);
				throw new InternalCompilerException("Constant folding produced errors!");
			} catch (IOException e) {
				throw new InternalCompilerException("Unexpected I/O error!", e);
			}
		}
	}
	
	public static void cleanupIfCJump(CFGNode node, IRNode before, IRNode after) {
		if (before instanceof IRCJump && after instanceof IRJump) {
			if (node.outEdges().size() != 2) 
				throw new InternalCompilerException("Invalid CJump node");
			/* Remove the dead edge from the CFG */
			String targetName = ((IRName) ((IRJump) after).target()).name();
			CFGNode node1 = node.outEdges().get(0), node2 = node.outEdges().get(1);
			Object node1Val = node1.value(), node2Val = node2.value();
			if (node1Val instanceof String && node1Val.equals(targetName) 
				|| node1Val instanceof IRLabel && ((IRLabel) node1Val).name().equals(targetName))
				node.removeEdgeTo(node2);
			else if (node2Val instanceof String && node2Val.equals(targetName)
				|| node2Val instanceof IRLabel && ((IRLabel) node2Val).name().equals(targetName))
				node.removeEdgeTo(node1);
			else if (targetName == null) {
				IRCJump oldJump = (IRCJump) before;
				String removeName = oldJump.trueLabel() != null ? oldJump.trueLabel() : oldJump.falseLabel();
				if (node1Val instanceof String && node1Val.equals(removeName) 
						|| node1Val instanceof IRLabel && ((IRLabel) node1Val).name().equals(removeName))
					node.removeEdgeTo(node1);
				else if (node2Val instanceof String && node2Val.equals(removeName)
					|| node2Val instanceof IRLabel && ((IRLabel) node2Val).name().equals(removeName))
					node.removeEdgeTo(node2);
				else
					throw new InternalCompilerException("Could not find dead edge");
				
				node.removeSelf();
			} else {
				throw new InternalCompilerException("Could not find dead edge");
			}
		}
	}
}
