package xic.analysis;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import xic.analysis.DataFlowAnalysis.Direction;
import xic.analysis.DataFlowAnalysis.FlowInfo;
import xic.ir.NameGenerator;


/**
 * Produces GraphViz code for visualizing a CFG. Can optionally
 * annotate the CFG with the results from a data flow analysis
 *
 */
public class DotGenerator {
	private Map<CFGNode, String> seen;
	private Map<CFGNode, String> in;
	private Map<CFGNode, String> out;
	private NameGenerator nameGen;

	public DotGenerator() {
		this.seen = new HashMap<CFGNode, String>();
		this.nameGen = new NameGenerator("N");
		this.in = new HashMap<CFGNode, String>();
		this.out = new HashMap<CFGNode, String>();
	}

	public DotGenerator(NameGenerator nameGen) {
		this.seen = new HashMap<CFGNode, String>();
		this.nameGen = nameGen;
		this.in = new HashMap<CFGNode, String>();
		this.out = new HashMap<CFGNode, String>();
	}
	
	public Map<CFGNode, String> getIn() {
		return in;
	}

	public void setIn(Map<CFGNode, String> in) {
		this.in = in;
	}
	
	public Map<CFGNode, String> getOut() {
		return out;
	}

	public void setOut(Map<CFGNode, String> out) {
		this.out = out;
	}

	public static String toDot(Map<String, CFGNode> cfg) {
		return new DotGenerator().generate(cfg);
	}
	
	public void setDataFlowAnalysis(DataFlowAnalysis dfa, Set<CFGNode> nodes) {
		in.clear();
		out.clear();
		for (CFGNode cfgNode : nodes) {
			FlowInfo fi = dfa.getFlowInfo(cfgNode);
			if (fi == null)
				continue;
			if (dfa.getDirection() == Direction.FORWARD)
				in.put(cfgNode, fi.entry.toString());
			else
				in.put(cfgNode, fi.exit.toString());
			if (dfa.getDirection() == Direction.FORWARD)
				out.put(cfgNode, fi.exit.toString());
			else
				out.put(cfgNode, fi.entry.toString());
		}
	}
	private static String shortClassName(Object o) {
		return o.getClass().getSimpleName();
	}
	public void setMultiDataFlowAnalysis(List<DataFlowAnalysis> dfas, Set<CFGNode> nodes) {
		//in.clear();
		//out.clear();
		for(DataFlowAnalysis dfa : dfas)
			for (CFGNode cfgNode : nodes) {
				FlowInfo fi = dfa.getFlowInfo(cfgNode);
				if (fi == null)
					continue;
				String in_head = "";
				String out_head = "";

				if(in.containsKey(cfgNode))
					in_head = in.get(cfgNode) + "\n";
				if(out.containsKey(cfgNode))
					out_head = out.get(cfgNode) + "\n";
				
				if (dfa.getDirection() == Direction.FORWARD)
					in.put(cfgNode, in_head + shortClassName(dfa) + ":" 
							+ fi.entry.toString());
				else
					in.put(cfgNode, in_head + shortClassName(dfa) + ":"  
							+ fi.exit.toString());
				if (dfa.getDirection() == Direction.FORWARD)
					out.put(cfgNode, out_head + shortClassName(dfa)  + ":" 
							+ fi.exit.toString());
				else
					out.put(cfgNode, out_head + shortClassName(dfa)  + ":" 
							+ fi.entry.toString());
			}
	}

	public String generate(Map<String, CFGNode> cfgs) {
		StringBuffer buf = new StringBuffer(
				"digraph dot_representation {\n" +
				"\trankdir = TB;\n" +
				"\tsize = \"8,5\"\n" +
				"\tnode [shape = rectangle];\n");

		for (String functionName : cfgs.keySet()) {
			CFGNode node = cfgs.get(functionName);
			Set<CFGNode> nodes = node.flatten();
			for (CFGNode cfgNode : nodes)
				buf.append(generateDotNode(cfgNode));
		}
		buf.append("}");
		return buf.toString();
	}
	
	public String generate(Set<CFGNode> cfgNodes) {
		StringBuffer buf = new StringBuffer(
				"digraph dot_representation {\n" +
				"\trankdir = TB;\n" +
				"\tsize = \"8,5\"\n" +
				"\tnode [shape = rectangle];\n");

		for (CFGNode cfgNode : cfgNodes)
			buf.append(generateDotNode(cfgNode));
		buf.append("}");
		return buf.toString();
	}

	public String getName(CFGNode n) {
		return seen.get(n);
	}

	public String prepareString(String s, boolean html) {
		if (html) {
			return s.trim()
				.replaceAll("\\r", "")
				.replaceAll("&", "&amp;")
				.replaceAll("<", "&lt;")
				.replaceAll(">", "&gt;")
				.replaceAll("\\n", "<br/>");
		} else {
			return s.trim().replaceAll("\\n", "\\\\n")
				.replaceAll("\\r", "");
		}
	}

	public String generateDotNode(CFGNode cfgNode) {
		StringBuffer buf = new StringBuffer();
		String dotName = null;
		
		if (seen.containsKey(cfgNode)) {
			dotName = seen.get(cfgNode);
		} else {
			dotName = nameGen.nextLabel();
			seen.put(cfgNode, dotName);
		}

		if (in.containsKey(cfgNode) || out.containsKey(cfgNode)) {
			buf.append("\t" + dotName + " [label = <");
			buf.append("<table border=\"0\">");
			buf.append("<tr><td bgcolor=\"gray\" align=\"text\">");
			if (in.containsKey(cfgNode))
				buf.append(prepareString(in.get(cfgNode), true));
			buf.append("</td></tr>");
			buf.append("<tr><td></td></tr>");
			buf.append("<tr><td>");
			buf.append(prepareString(cfgNode.toString(), true));
			buf.append("</td></tr>");
			buf.append("<tr><td></td></tr>");
			buf.append("<tr><td bgcolor=\"gray\" align=\"text\">");
			if (out.containsKey(cfgNode))
				buf.append(prepareString(out.get(cfgNode), true));
			buf.append("</td></tr>");
			buf.append("</table>");
			buf.append(">];\n");
		} else {
			buf.append("\t" + dotName + " [label = \""
					+ prepareString(cfgNode.toString(), false)
					+ "\"];\n");
		}

		for(CFGNode childCFGNode: cfgNode.outEdges()) {
			String childDotName;

			if(seen.containsKey(childCFGNode)) {
				childDotName = seen.get(childCFGNode);
			} else {
				childDotName = nameGen.nextLabel();
				seen.put(childCFGNode, childDotName);
			}
			buf.append("\t"+dotName + " -> " + childDotName + ";\n");
		}
		return buf.toString();
	}
}
