package xic.analysis;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import xic.InternalCompilerException;
import xic.codegen.RegAlloc;
import xic.ir.IRCJump;
import xic.ir.IRConst;
import xic.ir.IRDataSeq;
import xic.ir.IRDecl;
import xic.ir.IRExp;
import xic.ir.IRExpr;
import xic.ir.IRJump;
import xic.ir.IRLabel;
import xic.ir.IRMem;
import xic.ir.IRMove;
import xic.ir.IRNode;
import xic.ir.IRSeq;
import xic.ir.IRStmt;
import xic.ir.IRTemp;
import xic.ir.StructHashSet;


public class PropagatedExpressions extends DataFlowAnalysis {
	HashMap<String, StructHashSet<IRNode>> dep = new HashMap<String,StructHashSet<IRNode>>();
	public static class PEPair {
		String temp;
		IRExpr expr;

		public PEPair(String temp, IRExpr expr) {
			this.temp = temp;
			this.expr = expr;
		}

		public boolean equals(Object obj) {
			if (obj instanceof PEPair) {
				PEPair v = (PEPair) obj;
				return (temp.equals(v.temp) && expr.equalsTree(v.expr));
			}
			else return false;
		}

		public int hashCode() {
			return 17 + temp.hashCode() + expr.hashCode() * 37;
		}
		
		public String toString() {
			return temp + " -> " + expr;
		}
	};

	/// Value class used during copy propagation analysis
	public static class PEValue implements Value {
		/* List of variable pairs */
		private Set<PEPair> pairs;

		public PEValue() {
			pairs = new HashSet<PEPair>();
		}
		
		public PEValue(Set<PEPair> pairs) {
			this.pairs = new HashSet<PEPair>(pairs);
		}

		public Value copy() {
			return new PEValue(pairs);
		}

		public void meet(Value v) {
			pairs.retainAll(((PEValue) v).pairs);
		}
		
		public String toString() {
			return pairs.toString();
		}

		public int hashCode() {
			return pairs.hashCode();
		}
		
		public Set<PEPair> pairs() {
			return pairs;
		}
		public Set<String> temps() {
			Set<String> temps = new HashSet<String>();
			for(PEPair p : pairs)
				temps.add(p.temp);
			
			return temps;
		}
		public Set<IRExpr> exprs() {
			Set<IRExpr> exprs = new HashSet<IRExpr>();
			for(PEPair p : pairs)
				exprs.add(p.expr);
			
			return exprs;
		}

		public Map<String,IRExpr> uniquePairs() {
			Map<String,ArrayList<PEPair>> temp2pairs = new HashMap<String,ArrayList<PEPair>>();
			Map<String,IRExpr> uniq = new HashMap<String,IRExpr>();
			for(PEPair p : pairs) {
				if(temp2pairs.get(p.temp) == null)
					temp2pairs.put(p.temp, new ArrayList<PEPair>());
				temp2pairs.get(p.temp).add(p);
			}
			for(String temp : temp2pairs.keySet())
				if(temp2pairs.get(temp).size()==1)
					uniq.put(temp,temp2pairs.get(temp).get(0).expr);
					
			return uniq;
		}
		
		public boolean equals(Object obj) {
			if (!(obj instanceof PEValue))
				return false;
			return this.pairs.equals(((PEValue) obj).pairs);
		}

		public void setNullMeet() {
			pairs.clear();
		}
	};

	/// Dataflow-related information associated with each node
	public class PEFlowInfo extends FlowInfo {
		public Set<PEPair> gen;
		public Set<String> kill;

		public PEFlowInfo(Object o) {
			CFGNode node = (CFGNode)o;
			// for kill, actually don't need to store every possible pair,
			// only need one variable
			gen = new HashSet<PEPair>();
			kill = new HashSet<String>();

			entry = new PEValue();
			exit = new PEValue();
			computeCopyKill(node, gen, kill);
		}
	};

	public PropagatedExpressions(Set<CFGNode> nodes) {
		super(nodes);
		dep = new HashMap<String,StructHashSet<IRNode>>();

		dir = Direction.FORWARD;
		Set<PEPair> allPairs = new HashSet<PEPair>(); 
		for (CFGNode node : nodes)
			computeCopyKill(node, allPairs, null);
		top = new PEValue(allPairs);
	}

	protected FlowInfo createFlowInfo(CFGNode node) {
		return new PEFlowInfo(node);
	}

	public PEValue in(CFGNode node) {
		PEFlowInfo info = (PEFlowInfo) flowInfo.get(node);
		return (PEValue) info.entry;
	}
	
	public PEValue out(CFGNode node) {
		PEFlowInfo info = (PEFlowInfo) flowInfo.get(node);
		return (PEValue) info.exit;
	}

	private void computeCopyKill(CFGNode node, Set<PEPair> copy, Set<String> kill) {
		if (node.value() instanceof IRStmt) {
			IRStmt irStmt = (IRStmt) node.value();

			if(irStmt.containsCalls() && kill != null) {
				kill.add(RegAlloc.CALL_ARG_SCRATCH);
				kill.add(RegAlloc.CALL_RV_SCRATCH);
				kill.add(RegAlloc.MEM_NAME);
			}

			if (irStmt instanceof IRMove) {
				IRMove move = (IRMove) irStmt;
				if (move.target() instanceof IRTemp) {
					String targetName = ((IRTemp) move.target()).name();

					// x = e [where e contains no calls]
					if (!move.expr().containsCalls() 
							&& !(move.expr() instanceof IRTemp)
							&& !(move.expr() instanceof IRConst)) {
						
						Set<String> temps = new HashSet<String>();
						move.expr().findReadTemps(temps);

						if (dep!=null) {	
							for(String temp : temps) {
								if(dep.containsKey(temp))
									dep.get(temp).add(move.expr());
								else {
									dep.put(temp, new StructHashSet<IRNode>());
									dep.get(temp).add(move.expr());
								}
							}
						}
						copy.add(new PEPair(targetName, move.expr()));
						
						if (kill != null)
							kill.add(targetName);
						
					} else {
						if (kill != null) {
							// x = e [where e contains call or temp]
							kill.add(targetName);
						}
					}
				} else if(move.target() instanceof IRMem) {
					if (kill != null) {
						kill.add(AvailableExpressions.victim(move.target()));
					}
				}
					
			} else if(irStmt instanceof IRDataSeq
						|| irStmt instanceof IRJump
						|| irStmt instanceof IRLabel	
						|| irStmt instanceof IRSeq
						|| irStmt instanceof IRDecl
						|| irStmt instanceof IRExp
						|| irStmt instanceof IRCJump) {
				// don't do anything
			} else {
				throw new InternalCompilerException("Unknown IRStmt: "+irStmt);
			}
		}
	}

	// transform function: out(n) = gen(n) cup (in(n) - kill (n))
	protected void F(FlowInfo f) {
		
		PEFlowInfo fi = (PEFlowInfo) f;
		Set<PEPair> entry = ((PEValue) fi.entry).pairs();
		Set<PEPair> exit = ((PEValue) fi.exit).pairs();
	
		exit.clear();
		exit.addAll(entry);
		Iterator<PEPair> it = exit.iterator();
		while (it.hasNext()) {
			PEPair pair = it.next();
			if (fi.kill.contains(pair.temp)) {
				it.remove();
				continue;
			}
			for( String k : fi.kill) {
				if(dep.containsKey(k) 
					&& dep.get(k).contains(pair.expr)) {
					it.remove();
					break;
				}
			}
		}
		exit.addAll(fi.gen);
	}
}
