package xic.ast;

import java.util.ArrayList;
import java.util.List;

import xic.InternalCompilerException;
import xic.ir.NameMangler;

/**
 * Instances of DataSource describe whether an expression is stored using
 * temporaries, or if it is an input parameter to a function. Information about
 * the temporary names / function parameter offsets are also recorded.
 */
public class DataSource {
	public enum Type {
		GLOBAL,
		TEMP,
		PARAM,
		ATTRIBUTE
	};

	public static class Entry {
		private Type type;
		private String name;
		private int offset;
		private TypeNode tn;

		public Entry(Type type, String name, int offset) {
			this.type = type;
			this.name = name;
			this.offset = offset;
		}

		public Entry(Type type, String name) {
			this.type = type;
			this.name = name;
		}

		public Entry(Type type, int offset) {
			this.type = type;
			this.offset = offset;
		}
		
		public void setTypeNode(TypeNode tn) {
			this.tn = tn;
		}
	
		public boolean isTemp() {
			return type == Type.TEMP;
		}

		public boolean isGlobal() {
			return type == Type.GLOBAL;
		}

		public boolean isAttribute() {
			return type == Type.ATTRIBUTE;
		}

		public boolean isParam() {
			return type == Type.PARAM;
		}
		
		public int offset() {
			return offset;
		}
		
		public String name() {
			if (isGlobal())
				return NameMangler.translateGlobalVarName(name, tn);
			else
				return name;
		}

		public String toString() {
			return "ds[type="+type+", name="+name+", offset="+offset+"]";
		}
	};

	private List<Entry> entries;

	public DataSource() {
		entries = new ArrayList<Entry>();
	}

	public int size() {
		return entries.size();
	}

	public void addAttribute(String name, List<Integer> offsets) {
		for (int i: offsets)
			entries.add(new Entry(Type.ATTRIBUTE, name, i));
	}

	public void addGlobalName(String name, TypeNode typeNode, int offset) {
		Entry entry = new Entry(Type.GLOBAL, name, offset);
		entry.setTypeNode(typeNode);
		entries.add(entry);
	}

	public void addTempName(String name) {
		entries.add(new Entry(Type.TEMP, name));
	}

	public void addParamOffset(int tempOffset) {
		entries.add(new Entry(Type.PARAM, tempOffset));
	}

	public void addParamOffset(int tempOffset, String name) {
		entries.add(new Entry(Type.PARAM, name, tempOffset));
	}

	public void add(DataSource ds) {
		if (ds == null)
			return;
		entries.addAll(ds.entries);
	}

	public void add(Entry entry) {
		entries.add(entry);
	}

	public Entry get(int index) {
		return entries.get(index);
	}

	public String getTempName(int index) {
		if (!isTemp(index))
			throw new InternalCompilerException("Parameter " + index + " is not a temporary");
		return entries.get(index).name;
	}

	public int getParamOffset(int index) {
		if (!isParam(index))
			throw new InternalCompilerException("Parameter " + index + " is not a temporary");
		return entries.get(index).offset;
	}


	public boolean isParam(int entry) {
		return entries.get(entry).name == null;
	}

	public boolean isTemp(int entry) {
		return entries.get(entry).name != null;
	}

	public String toString() {
		return entries.toString();
	}

	public DataSource subList(int a, int b) {
		DataSource result = new DataSource();
		result.entries.addAll(entries.subList(a, b));
		return result;
	}
}
