package xic.ir;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import xic.ast.ASTNode;
import xic.ast.Copyable;
import xic.visitors.NodeVisitor;

import edu.cornell.cs.cs4120.util.VisualizableTreeNode;

public class IRSeq extends IRStmt {
	private List<IRStmt> stmts;

	public IRSeq(List<IRStmt> stmts) {
		this.stmts = stmts;
	}

	public IRSeq() {
		stmts = new ArrayList<IRStmt>();
	}

	public void appendStmt(IRStmt stmt) {
		if (stmt != null) {
			/* Flatten nested SEQs here already */
			if (stmt instanceof IRSeq)
				stmts.addAll(((IRSeq) stmt).stmts());
			else
				stmts.add(stmt);
		}
	}

	
	public List<IRStmt> stmts() {
		return stmts;
	}

    public ASTNode visitChildren(NodeVisitor v) {
    	boolean modified = false;

    	ArrayList<IRStmt> results = new ArrayList<IRStmt>();
    	for (IRStmt stmt : stmts) {
    		IRStmt newStmt = (IRStmt) v.visit(this, stmt);
    		if (newStmt != stmt)
    			modified = true;
    		if (newStmt instanceof IRSeq)
    			results.addAll(((IRSeq) newStmt).stmts);
    		else
    			results.add(newStmt);
    	}

		if (modified)
			return new IRSeq(results);

		return this;
	}
    
    public Copyable copy() {
		return new IRSeq(new ArrayList<IRStmt>(stmts));
    }

    public Copyable deepCopy() {
    	List<IRStmt> stmtsCopy = new ArrayList<IRStmt>();
    	for (IRStmt s : this.stmts)
    		stmtsCopy.add((IRStmt) s.deepCopy());
		return new IRSeq(stmtsCopy);
    }

    public boolean containsCalls() {
    	boolean result = false;
    	for (IRStmt stmt : stmts)
    		result |= stmt.containsCalls();
    	return result;
    }

    public boolean containsMemReads() {
    	boolean result = false;
    	for (IRStmt stmt : stmts)
    		result |= stmt.containsMemReads();
    	return result;
    }

    public boolean containsMemWrites() {
    	boolean result = false;
    	for (IRStmt stmt : stmts)
    		result |= stmt.containsMemWrites();
    	return result;
    }

    public void findReadTemps(Set<String> tempNames) {
    	for (IRStmt stmt : stmts)
    		stmt.findReadTemps(tempNames);
    }

    public void findWrittenTemps(Set<String> tempNames) {
    	for (IRStmt stmt : stmts)
    		stmt.findWrittenTemps(tempNames);
    }
   
	public int nodeCount() {
		int nodeCount = 0;
		for (IRNode n : stmts)
			nodeCount += n.nodeCount();
		return nodeCount + 1;
	}

	public int computeMaximumCallArguments() {
		int value = 0;
		for (IRStmt s : stmts)
			value = Math.max(value, s.computeMaximumCallArguments());
		return value;
	}

	public int computeMaximumCallResults() {
		int value = 0;
		for (IRStmt s : stmts)
			value = Math.max(value, s.computeMaximumCallResults());
		return value;
	}
	public Iterable<VisualizableTreeNode> children() {
		ArrayList<VisualizableTreeNode> result 
			= new ArrayList<VisualizableTreeNode>(stmts.size());
		result.addAll(stmts);
		return result;
	}


	public String label() {
		return "SEQ";
	}
	
	public boolean equalsTree(Object object) {
		if (!(object instanceof IRSeq))
			return false;
		IRSeq other = (IRSeq) object;
		if (other.stmts.size() != stmts.size())
			return false;
		
		for (int i=0; i<stmts.size(); ++i)
			if (!stmts.get(i).equalsTree(other.stmts.get(i)))
				return false;
		return true;
	}
	
	public int treeHashCode() {
		int code=17;
		for(IRStmt stmt : stmts)
			code= 37*code + stmt.treeHashCode();
		return code;
	}

}
