package xic.lexer;

import edu.cornell.cs.cs4120.xi.CompilationException;
import java.util.NoSuchElementException;
import java.util.Iterator;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import xic.PositionImpl;
import xic.parser.TokenType;
import java_cup.runtime.Symbol;

@SuppressWarnings("unused")

%%

%{ 
	private String unit;
	private StringBuffer string = new StringBuffer();
	private StringBuffer srcString = new StringBuffer();
	private int stringLine, stringCol;
	private HashMap<String, Integer> keywords = new HashMap<String, Integer>();
	private BaseToken cachedToken = null;
	private int firstToken;

	public XiLexer(String unit, int firstToken, java.io.Reader in) {
		this(in);
		this.unit = unit;
		this.firstToken = firstToken;
		
		keywords.put("if", TokenType.IF);
		keywords.put("else", TokenType.ELSE);
		keywords.put("while", TokenType.WHILE);
		keywords.put("break", TokenType.BREAK);
		keywords.put("return", TokenType.RETURN);
		keywords.put("use", TokenType.USE);
		keywords.put("length", TokenType.LENGTH);
		keywords.put("int", TokenType.INT);
		keywords.put("bool", TokenType.BOOL);
		keywords.put("true", TokenType.TRUE);
		keywords.put("false", TokenType.FALSE);
		keywords.put("class", TokenType.CLASS);
		keywords.put("new", TokenType.NEW);
		keywords.put("extends", TokenType.EXTENDS);
		keywords.put("this", TokenType.THIS);
		keywords.put("null", TokenType.NULL);
	}

	public BaseToken next() throws CompilationException {
		if (cachedToken != null) {
			BaseToken result = cachedToken;
			cachedToken = null;
			return result;
		}

		try {
			return (BaseToken)next_token().value;
		} catch (IOException e) {
			throw new NoSuchElementException("I/O Error");
		}
	}

	Symbol makeToken(int type) {
		return new Symbol(type, new BaseToken(type, yytext(), unit, yyline, yycolumn, yylength()));
	}

	public void remove() {
		throw new UnsupportedOperationException();
	}

	public boolean hasNext() {
		if (cachedToken != null)
			return true;
		try {
			Symbol token = next_token(); // Throws when running out of tokens
			if (token.sym == TokenType.EOF)
				return false;
			cachedToken = (BaseToken) token.value;
			return true;
		} catch (IOException e) {
			throw new NoSuchElementException("I/O Error");
		}
	}

	public static void main(String args[]) {
		XiLexer lexer = new XiLexer("<stdin>", TokenType.SOURCE_IDENTIFIER, new InputStreamReader(System.in));
		while (true)
			System.out.println(lexer.next().toString());
	}
%}

%line
%column
%full
%cup
%unicode
%implements Iterator<BaseToken>
%class XiLexer
%public

%eofval{
	return new Symbol(TokenType.EOF,new BaseToken(TokenType.EOF, yytext(), unit, yyline, yycolumn, yylength())); 
%eofval}


ALPHA=[A-Za-z]
ESCAPED_CHARACTER_LITERAL = '\\.'
CHARACTER_LITERAL='[^']'
WHITESPACE = [\n\r\t ]
IDENT = {ALPHA}({ALPHA}|{DIGIT}|_|')*
COMMENT = \/\/.*
DIGIT=[0-9]
NUMBER={DIGIT}+
NEG_MAX=-9223372036854775808

%state STRING
%xstate MAIN

%% 

<STRING> {
  \\\"                           { string.append('"'); srcString.append(yytext()); }
  \"                             { yybegin(MAIN);
                                   srcString.append(yytext());
                                   return new Symbol(TokenType.STRING_LITERAL,
														new StringToken(TokenType.STRING_LITERAL, 
                                   						  new PositionImpl(unit, stringLine, yyline, stringCol, yycolumn),
                                   						  srcString.toString(),
                                   						  string.toString()));
                                 }
  [^\n\r\"\\]+                   { string.append(yytext()); srcString.append(yytext());}
  \\t                            { string.append('\t'); srcString.append(yytext()); }
  \\n                            { string.append('\n'); srcString.append(yytext()); }

  \\r                            { string.append('\r'); srcString.append(yytext()); }
  
  \\\\                           { string.append('\\'); srcString.append(yytext()); }

  \\.  { 
  		throw new CompilationException("Encountered an illegal escape sequence: '" + yytext() + "'", 
			new PositionImpl(unit, yyline, yyline, yycolumn, yycolumn)); }
  .|\n|\r {
		throw new CompilationException("Encountered an illegal string character: '" + yytext() + "'", 
			new PositionImpl(unit, yyline, yyline, yycolumn, yycolumn));
  }
}

<YYINITIAL> {
	. {
		/* Return a special token, which identifies the type of source file */ 
		yybegin(MAIN);
		yypushback(1); 
		return new Symbol(firstToken, new BaseToken(firstToken, null, unit, -1, -1, 0));
	}
	{WHITESPACE}+ {}
}

<MAIN> {  
  {COMMENT} {}

  \"        { stringLine = yyline;
              stringCol = yycolumn;
              string.setLength(0);
              srcString.setLength(0);
              srcString.append('"'); 
              yybegin(STRING);
            }

{NEG_MAX}   { return new Symbol(TokenType.INTEGER_LITERAL, new IntegerToken(TokenType.INTEGER_LITERAL, yytext(), unit, yyline, yycolumn, yylength())); }
{NUMBER}    { return new Symbol(TokenType.INTEGER_LITERAL, new IntegerToken(TokenType.INTEGER_LITERAL, yytext(), unit, yyline, yycolumn, yylength())); }
  "=="      { return makeToken(TokenType.EQUAL); }
  ">="      { return makeToken(TokenType.GEQ); }
  "<="      { return makeToken(TokenType.LEQ); }
  "!="      { return makeToken(TokenType.NOT_EQUAL); }
  "+"       { return makeToken(TokenType.PLUS); }
  "-"       { return makeToken(TokenType.MINUS); }
  "*"       { return makeToken(TokenType.TIMES); }
  "/"       { return makeToken(TokenType.DIVIDE); }
  "="       { return makeToken(TokenType.GETS); }
  "%"       { return makeToken(TokenType.MODULO); }
  "!"       { return makeToken(TokenType.NOT); }
  "<"       { return makeToken(TokenType.LT); }
  ">"       { return makeToken(TokenType.GT); }
  "&"       { return makeToken(TokenType.AND); }
  "|"       { return makeToken(TokenType.OR); }
  "["       { return makeToken(TokenType.OPEN_BRACKET); }
  "]"       { return makeToken(TokenType.CLOSE_BRACKET); }
  "("       { return makeToken(TokenType.OPEN_PAREN); }
  ")"       { return makeToken(TokenType.CLOSE_PAREN); }
  "{"       { return makeToken(TokenType.OPEN_BRACE); }
  "}"       { return makeToken(TokenType.CLOSE_BRACE); }
  ":"       { return makeToken(TokenType.COLON); }
  ","       { return makeToken(TokenType.COMMA); }
  ";"       { return makeToken(TokenType.SEMICOLON); }
  "_"       { return makeToken(TokenType.UNDERSCORE); }
  "."       { return makeToken(TokenType.DOT); }
  

  {IDENT}   {
  	String text = yytext();
  	if (keywords.containsKey(text))
  		return new Symbol(keywords.get(text), new BaseToken(keywords.get(text), text, unit, yyline, yycolumn, yylength()));
  	else 
  		return new Symbol(TokenType.IDENTIFIER, new BaseToken(TokenType.IDENTIFIER, text, unit, yyline, yycolumn, yylength()));
  }

  {ESCAPED_CHARACTER_LITERAL} { 
  	char unescaped;
  	String text = yytext();
  	/* 
		\n A new line 
		\r A carriage return 
		\t A horizontal tab 
		\\ A backslash (\)
		\' A single quote (')
  	*/
  	switch(text.charAt(2)) {
  		case 'n':
  			unescaped = '\n';
  			break;
  		case 'r':
  			unescaped = '\r';
  			break;
   		case 't':
  			unescaped = '\t';
  			break;
   		case '\\':
  			unescaped = '\\';
  			break;
   		case '\'':
  			unescaped = '\'';
  			break;
  		default:
	  		throw new CompilationException("Encountered an illegal escape sequence: '" + yytext() + "'", 
				new PositionImpl(unit, yyline, yyline, yycolumn, yycolumn));
    }
    return new Symbol(TokenType.CHARACTER_LITERAL, new CharacterToken(TokenType.CHARACTER_LITERAL, text, unescaped, unit, yyline, yycolumn, yylength()));
  }    	 		 
  
  {CHARACTER_LITERAL} { return new Symbol(TokenType.CHARACTER_LITERAL, new CharacterToken(TokenType.CHARACTER_LITERAL, yytext(), yytext().charAt(1), unit, yyline, yycolumn, yylength())); }

  {WHITESPACE}+ {}

. {
	throw new CompilationException("Encountered an illegal character: '" + yytext() + "'", 
		new PositionImpl(unit, yyline, yyline, yycolumn, yycolumn));
  }
}

